local S = superquest.S

local form_name = core.get_current_modname()..":flag_settings"
local node_positions = {}

local autotr_coords_field_hint = S("Should contain positive integer values of distance of x,y,z, relative to the position of the flag, separated by commas") .. ".\n" ..
                                 S("The values in the first field are subtracted from the flag's position while the values in the second field are added to it") .. ".\n" ..
                                 S("This flag will be automatically triggered for players detected within the cube described by the coordinates") .. "."

local teleport_coords_field_hint = S("Should contain integer values of distance of x,y,z, relative to the position of the flag, separated by commas") .. ".\n" ..
                                   S("The value can be empty that means that teleportation is disabled for the flag") .. "."

local function get_formspec(owner, network, name, from_dev, to_dev, teleport_coords, digiline_channel, flags_before, can_edit_owner)
    local height_cur = 0.5
    local formspec_data = ""

    if can_edit_owner then
        formspec_data = formspec_data .. "field[0.5,"..tostring(height_cur)..";5.0,0.8;owner;" .. S("Owner") .. ";"..core.formspec_escape(owner).."]"
        height_cur = height_cur + 1.3
    else
        formspec_data = formspec_data .. "label[0.5,"..tostring(height_cur)..";"..core.formspec_escape(S("Owner") .. ": "..owner).."]"
        height_cur = height_cur + 0.8
    end

    formspec_data = formspec_data ..
        "field[0.5,"..tostring(height_cur)..";5.0,0.8;network;" .. S("Network") .. ";"..core.formspec_escape(network).."]"
    height_cur = height_cur + 1.3

    formspec_data = formspec_data ..
        "field[0.5,"..tostring(height_cur)..";5.0,0.8;name;" .. S("Name") .. " ("..S("optional")..");"..core.formspec_escape(name).."]"
    height_cur = height_cur + 1.3

    if superquest.config.flags_autotrigger then
        formspec_data = formspec_data .. "field[0.5,"..tostring(height_cur)..";2.2,0.8;from_coords;" .. S("From coords") .. ";"..core.formspec_escape(from_dev).."]" ..
        "field[3.3,"..tostring(height_cur)..";2.2,0.8;to_coords;" .. S("To coords") .. ";"..core.formspec_escape(to_dev).."]" ..
        "tooltip[0.5,"..tostring(height_cur)..";5,0.8;"..autotr_coords_field_hint.."]"
        height_cur = height_cur + 1.2
    end

    if superquest.config.teleportation then
        formspec_data = formspec_data ..
            "field[0.5,"..tostring(height_cur)..";5,0.8;teleport_to;" .. S("Teleportation coords (relative)") .. ";"..core.formspec_escape(teleport_coords).."]" ..
            "tooltip[0.5,"..tostring(height_cur)..";5,0.8;"..teleport_coords_field_hint.."]"
        height_cur = height_cur + 1.2
    end

    if digilines then
        formspec_data = formspec_data ..
            "field[0.5,"..tostring(height_cur)..";4.2,0.8;digiline_channel;"..S("Digiline channel").." ("..S("optional")..");"..core.formspec_escape(digiline_channel).."]" ..
            "button[4.7,"..tostring(height_cur)..";0.8,0.8;digiline_guide;?]"
        height_cur = height_cur + 1.2
    end

    formspec_data = formspec_data ..
        "textarea[0.5,"..tostring(height_cur)..";5.0,2.8;flags_before;" .. S("Preceding flags") .. ";"..core.formspec_escape(flags_before).."]"
    height_cur = height_cur + 3.3

    formspec_data = formspec_data ..
        "button_exit[0.5,"..tostring(height_cur)..";5.0,0.8;save;" .. S("Save") .. "]"
    height_cur = height_cur + 1

    formspec_data = "formspec_version[4]" ..
        "size[6,"..tostring(height_cur).."]" ..
        formspec_data

    return formspec_data
end

local function get_network_index(owner, network)
    return owner .. "#" .. network
end

local get_flag_index = superquest.storage.coords_to_key

-- For digiline-requiring flags
local allowed_players = {}

local function allowed_players_add(owner, network, pos, player_name)
    local network_index = get_network_index(owner, network)
    local flag_index = get_flag_index(pos)

    if not allowed_players[network_index] then
        allowed_players[network_index] = {}
    end

    if not allowed_players[network_index][flag_index] then
        allowed_players[network_index][flag_index] = {}
    end

    local allowed_players_flag = allowed_players[network_index][flag_index]

    for _, v in ipairs(allowed_players_flag) do
        if v == player_name then
            return
        end
    end

    table.insert(allowed_players_flag, player_name)
    if #allowed_players_flag > 10 then
        table.remove(allowed_players_flag, 1)
    end
end

local function _allowed_players_remove_by_indices(network_index, flag_index, player_index)
    table.remove(allowed_players[network_index][flag_index], player_index)
    if #allowed_players[network_index][flag_index] == 0 then
        allowed_players[network_index][flag_index] = nil
        if not next(allowed_players[network_index]) then
            allowed_players[network_index] = nil
        end
    end
end

local function allowed_players_remove_for_flag(owner, network, pos, player_name)
    local index = 0
    local network_index = get_network_index(owner, network)
    local flag_index = get_flag_index(pos)
    if allowed_players[network_index] and allowed_players[network_index][flag_index] then
        for k, v in ipairs(allowed_players[network_index][flag_index]) do
            if v == player_name then
                index = k
                break
            end
        end
    end

    if index ~= 0 then
        _allowed_players_remove_by_indices(network_index, flag_index, index)

        return true
    else
        return false
    end
end

local function allowed_players_remove_for_all(owner, network, player_name)
    local network_index = get_network_index(owner, network)

    if allowed_players[network_index] then
        for flag_index, flag_players in pairs(allowed_players[network_index]) do
            for i_player, v_player in ipairs(flag_players) do
                if player_name == v_player then
                    _allowed_players_remove_by_indices(network_index, flag_index, i_player)
                    break
                end
            end
        end
    end
end

local function digline_send_flag_punched_if_enabled(network_obj, pos, channel, player_name, result)
    if digilines and channel ~= "" then
        local reached, total = network_obj:get_flags_stats_for_player(player_name)

        superquest.digilines.send.flag_punched(pos, channel, player_name, result, reached, total)
    end
end

local function try_to_count_flag(pos, player, print_to_chat, is_digiline_requiring)
    local meta = core.get_meta(pos)

    local owner = meta:get_string("owner")
    local network = meta:get_string("network")
    local flag_name = meta:get_string("name")
    if flag_name == "" then
        flag_name = nil
    end
    local channel = meta:get_string("digiline_channel")

    local player_name = player:get_player_name()

    if owner == "" or network == "" then
        if print_to_chat then
            core.chat_send_player(player_name, S("This flag is not configured") .. "!")
        end

        return
    end

    local network_obj = assert(superquest.Network(owner, network))

    if (owner == player_name or superquest.privileges.can_edit_all(player_name))
            and not network_obj:is_in_user_mode() then
        if print_to_chat then
            core.chat_send_player(player_name, "You need to switch to the User mode to participate")
        end
        return
    end

    local timed_duration = network_obj:get_timed_duration()

    if timed_duration ~= 0 then
        local active_network = superquest.active_timed_quests:get_active_network_for_player(player_name)
        if not active_network or active_network.owner ~= owner or active_network.network ~= network then
            if print_to_chat then
                core.chat_send_player(player_name, S("This flag belongs to a timed quest. You need to start it first in the corresponding quest machine."))
                digline_send_flag_punched_if_enabled(network_obj, pos, channel, player_name, "not_started")
            end
            return
        end
    end

    local max_completions = network_obj:get_max_completions()

    if max_completions ~= 0 and network_obj:get_completions_for_player(player_name) >= max_completions then
        if print_to_chat then
            core.chat_send_player(player_name, S("You already reached the maximum number of completions"))
            digline_send_flag_punched_if_enabled(network_obj, pos, channel, player_name, "reached_max_completions")
        end
        return
    end

    local is_counted = network_obj:is_flag_counted(pos, flag_name, player_name)
    if is_counted then
        if print_to_chat then
            core.chat_send_player(player_name, S("This flag is already counted") .. "!")
            digline_send_flag_punched_if_enabled(network_obj, pos, channel, player_name, "already_counted")
        end
        return
    end

    if meta:contains("flags_before") then
        local flags_before = core.deserialize(meta:get_string("flags_before"))
        for _, v in ipairs(flags_before) do
            if network_obj:is_name_or_pos_available(nil, v) then
                if print_to_chat and core.get_player_by_name(owner) then
                    core.chat_send_player(owner, S("Warning: the network @1 doesn't have a flag with the name @2", network, v))
                end
            elseif not network_obj:is_flag_counted(nil, v, player_name) then
                if print_to_chat then
                    local print_str = S("You missed flags before the current one!")
                    if superquest.config.teleportation then
                        local can_teleport = superquest.teleportation.allow_tp_for_flag(player_name)
                        if can_teleport then
                            print_str = print_str .. " " .. S("You can type '/superquest teleport' to teleport to the last triggered flag.")
                        end
                    end
                    core.chat_send_player(player_name, print_str)
                    digline_send_flag_punched_if_enabled(network_obj, pos, channel, player_name, "missed_previous")
                end
                return
            end
        end
    end

    if is_digiline_requiring then
        local removed = allowed_players_remove_for_flag(owner, network, pos, player_name)
        if not removed then
            if print_to_chat then
                core.chat_send_player(player_name, S("You cannot trigger this flag without the corresponding digiline signal.").." "..
                        S("Receiving this signal depends on the quest itself. Ask the quest's owner if it's not obvious how to get it."))
                digline_send_flag_punched_if_enabled(network_obj, pos, channel, player_name, "no_digiline_signal")
            end
            return
        end
    end

    network_obj:add_player_reached_flag(pos, flag_name, player_name)

    local reached, total = network_obj:get_flags_stats_for_player(player_name)
    if reached >= total then
        if timed_duration then
            superquest.active_timed_quests:stop_and_remove(player_name)
        end

        if superquest.config.teleportation then
            if network_obj:get_teleport_after_compl() then
                superquest.teleportation.set_tp_dest_to_quest_machine(player_name, owner, network, false)
                core.chat_send_player(player_name, S("Quest completed").."! "..S("You can type '/superquest teleport' to teleport to the Quest Machine.")..
                        " "..S("Punch the Quest Machine to receive the reward."))
            end
        end
    elseif superquest.config.teleportation then
        superquest.teleportation.set_tp_dest_to_flag_if_possible(player_name, owner, network, pos)
    end

    digline_send_flag_punched_if_enabled(network_obj, pos, channel, player_name, "triggered")

    local maxpos_deviation_table = {
        [0] = vector.new(0, 0.5, 0),
              vector.new(0, 0, 0.5),
              vector.new(0, 0, -0.5),
              vector.new(0.5, 0, 0),
              vector.new(-0.5, 0, 0),
              vector.new(0, -0.5, 0),
    }
    local deviation = maxpos_deviation_table[math.floor(core.get_node(pos).param2 / 4)]

    core.add_particlespawner({
        amount = 50,
        time = 0.01,
        minpos = vector.add(pos, deviation * 2),
        maxpos = vector.add(pos, deviation * 3),
        minvel = vector.new(-2, -2, -2),
        maxvel = vector.new(2, 2, 2),
        minexptime = 0.3,
        maxexptime = 0.3,
        minsize = 0.4,
        maxsize = 0.4,
        texture = is_digiline_requiring and "superquest_particle_blue.png" or "superquest_particle_red.png"
    })

    superquest.hud.show_popup(player, S("Flags counted") .. ": "..reached.." / "..total)
end

local color_mark_entity_name = "superquest:color_mark"

core.register_entity(color_mark_entity_name, {
    initial_properties = {
        visual = "mesh",
        mesh = "superquest_color_mark.obj",
        textures = {
            "superquest_pixel.png",
        },
        pointable = false,
        static_save = false,
        visual_size = {x = 10, y = 10, z = 10},
    }
})

-- Haven't come up with a better solution yet
local rotation_vectors = {
    [0] = vector.new(0, 0, 0),
    [1] = vector.new(0, -math.pi/2, 0),
    [2] = vector.new(0, math.pi, 0),
    [3] = vector.new(0, math.pi/2, 0),
    [4] = vector.new(-math.pi/2, 0, 0),
    [5] = vector.new(0, -math.pi/2, -math.pi/2),
    [6] = vector.new(math.pi/2, 0, math.pi),
    [7] = vector.new(math.pi, -math.pi/2, -math.pi/2),
    [8] = vector.new(math.pi/2, 0, 0),
    [9] = vector.new(math.pi, math.pi/2, -math.pi/2),
    [10] = vector.new(-math.pi/2, 0, math.pi),
    [11] = vector.new(0, math.pi/2, -math.pi/2),
    [12] = vector.new(math.pi, math.pi, -math.pi/2),
    [13] = vector.new(-math.pi/2, 0, math.pi/2),
    [14] = vector.new(0, math.pi, -math.pi/2),
    [15] = vector.new(math.pi/2, 0, math.pi/2),
    [16] = vector.new(0, 0, -math.pi/2),
    [17] = vector.new(math.pi/2, 0, -math.pi/2),
    [18] = vector.new(math.pi, 0, -math.pi/2),
    [19] = vector.new(-math.pi/2, 0, -math.pi/2),
    [20] = vector.new(math.pi, math.pi, 0),
    [21] = vector.new(math.pi, -math.pi/2, 0),
    [22] = vector.new(math.pi, 0, 0),
    [23] = vector.new(math.pi, math.pi/2, 0),
}

local function remove_color_mark(pos)
    for obj in core.objects_inside_radius(pos, 0.1) do
        local ent = obj:get_luaentity()
        if ent and ent.name == color_mark_entity_name then
            obj:remove()
        end
    end
end

local function add_color_mark(pos, param2, color)
    local ent = core.add_entity(pos, color_mark_entity_name)
    if not ent then
        return
    end

    ent:set_rotation(rotation_vectors[param2])
    ent:set_texture_mod("^[colorize:" .. color .. ":255")
end

local function on_rotate(pos, node, user, mode, new_param2)
    for obj in core.objects_inside_radius(pos, 0.1) do
        local ent = obj:get_luaentity()
        if ent and ent.name == color_mark_entity_name then
            obj:set_rotation(rotation_vectors[new_param2])
        end
    end
end

local function flag_after_place(pos, placer, itemstack)
    local meta = core.get_meta(pos)
    if placer and placer:is_player() then
        meta:set_string("owner", placer:get_player_name())
    end
end

local function flag_on_destruct(pos)
    remove_color_mark(pos)

    local meta = core.get_meta(pos)

    local network_obj = superquest.Network(meta:get_string("owner"), meta:get_string("network"))
    if not network_obj then
        return
    end

    local flag_name = meta:get_string("name")
    if flag_name == "" then
        flag_name = nil
    end

    network_obj:remove_flag(pos, flag_name)

    -- We don't need to check the player because flags should be placed in protected areas anyway
end

local function flag_on_place(itemstack, placer, pointed_thing)
    local dir = core.dir_to_wallmounted(
        vector.subtract(pointed_thing.under, pointed_thing.above)
    )
    local facedir_table = {[0] = 20, 0, 17, 15, 8, 6}
    if dir == 0 then
        facedir_table[dir] = facedir_table[dir] + (4 - core.dir_to_facedir(placer:get_look_dir())) % 4
    elseif dir == 1 then
        facedir_table[dir] = facedir_table[dir] + core.dir_to_facedir(placer:get_look_dir())
    end

    return core.item_place(itemstack, placer, pointed_thing, facedir_table[dir])
end

local function show_menu(pos, player)
    if not (player and player:is_player()) then
        return
    end

    local meta = core.get_meta(pos)

    local player_name = player:get_player_name()

    if not (meta:get_string("owner") == player_name or superquest.privileges.can_edit_all(player_name)) then
        return
    end

    node_positions[player_name] = pos

    local from_v = meta:get_int("scan_from_x")..","..meta:get_int("scan_from_y")..","..meta:get_int("scan_from_z")
    local to_v = meta:get_int("scan_to_x")..","..meta:get_int("scan_to_y")..","..meta:get_int("scan_to_z")

    local flags_before_str = ""
    if meta:contains("flags_before") then
        local flags_before = core.deserialize(meta:get_string("flags_before"))
        flags_before_str = table.concat(flags_before, "\n")
    end

    core.show_formspec(player_name, form_name,
        get_formspec(meta:get_string("owner"), meta:get_string("network"), meta:get_string("name"),
        from_v, to_v, meta:get_string("teleport_to"), meta:get_string("digiline_channel"),
        flags_before_str, superquest.privileges.can_edit_all(player_name)))
end

core.register_node("superquest:flag", {
    description = S("Flag"),
    drawtype = "mesh",
    mesh = "superquest_flag.obj",
    tiles = { "superquest_flag.png" },
    inventory_image = "superquest_flag.png",
    selection_box = {
        type = "fixed",
        fixed = { -0.1, -0.5, 0.1, 0.1, 1.5, -0.5 }
    },
    paramtype = "light",
    paramtype2 = "facedir",
    sunlight_propagates = true,
    use_texture_alpha = "clip",
    walkable = false,
    groups = { cracky=1, oddly_breakable_by_hand = 2 },

    digilines = {
        receptor = {}
    },

    after_place_node = flag_after_place,

    on_rightclick = function(pos, node, clicker, itemstack, pointed_thing)
        show_menu(pos, clicker)
    end,

    on_destruct = flag_on_destruct,

    on_punch = function(pos, node, puncher)
        if not (puncher and puncher:is_player()) then
            return
        end

        try_to_count_flag(pos, puncher, true, false)
    end,

    on_place = flag_on_place,

    on_rotate = on_rotate,
})

if digilines then
    local digiline_handlers = {
        allow_player = function(pos, player_name)
            local meta = core.get_meta(pos)
            local owner = meta:get_string("owner")
            local network = meta:get_string("network")
            local flag_name = meta:get_string("name")
            if flag_name == "" then
                flag_name = nil
            end

            if owner == "" or network == "" then
                return
            end

            local network_obj = assert(superquest.Network(owner, network))

            if network_obj:get_timed_duration() ~= 0 then
                local active_network = superquest.active_timed_quests:get_active_network_for_player(player_name)
                if not active_network or active_network.owner ~= owner or active_network.network ~= network then
                    return
                end
            end

            local max_completions = network_obj:get_max_completions()
            if max_completions ~= 0 and network_obj:get_completions_for_player(player_name) >= max_completions then
                return
            end

            local is_counted = network_obj:is_flag_counted(pos, flag_name, player_name)
            if is_counted then
                return
            end

            allowed_players_add(owner, network, pos, player_name)
        end,

        revoke_allow_player = function(pos, player_name)
            local meta = core.get_meta(pos)
            local owner = meta:get_string("owner")
            local network = meta:get_string("network")

            if owner == "" or network == "" then
                return
            end

            allowed_players_remove_for_flag(owner, network, pos, player_name)
        end,
    }
    core.register_node("superquest:flag_digiline_requiring", {
        description = S("Digiline-Requiring Flag"),
        drawtype = "mesh",
        mesh = "superquest_flag.obj",
        tiles = { "superquest_flag_blue.png" },
        inventory_image = "superquest_flag_blue.png",
        selection_box = {
            type = "fixed",
            fixed = { -0.1, -0.5, 0.1, 0.1, 1.5, -0.5 }
        },
        paramtype = "light",
        paramtype2 = "facedir",
        sunlight_propagates = true,
        use_texture_alpha = "clip",
        walkable = false,
        groups = { cracky=1, oddly_breakable_by_hand = 2 },

        digilines = {
            receptor = {},
            effector = {
                action = function(pos, _, channel, msg)
                    local meta = core.get_meta(pos)
                    if (meta:get_string("digiline_channel") ~= channel) then
                        return
                    end

                    superquest.digilines.parse.flag_digiline_requiring(digiline_handlers, pos, msg)
                end
            }
        },

        after_place_node = flag_after_place,

        on_rightclick = function(pos, node, clicker, itemstack, pointed_thing)
            show_menu(pos, clicker)
        end,

        on_destruct = flag_on_destruct,

        on_punch = function(pos, node, puncher)
            if not (puncher and puncher:is_player()) then
                return
            end

            try_to_count_flag(pos, puncher, true, true)
        end,

        on_place = flag_on_place,

        on_rotate = on_rotate,
    })
end

core.register_on_player_receive_fields(function(player, formname, fields)
    if formname ~= form_name then
        return false
    end

    if not (player and player:is_player()) then
        return true
    end

    local player_name = player:get_player_name()

    local pos = node_positions[player_name]
    if pos == nil then
        return true
    end

    local meta = core.get_meta(pos)
    local owner = meta:get_string("owner")

    if owner ~= player_name and not superquest.privileges.can_edit_all(player_name) then
        return true
    end

    if fields.save then
        if not fields.network and not fields.name and not fields.flags_before then
            return true
        end

        local old_network = meta:get_string("network")
        local new_network = fields.network:gsub('^%s*(.-)%s*$', '%1')

        if new_network:len() > superquest.config.max_network_name_length then
            core.chat_send_player(player_name, S("Max network name length (@1 symbols) exceeded", superquest.config.max_network_name_length))
            return true
        end
        if new_network:find("#") then
            core.chat_send_player(player_name, S("Using the '#' symbol is not allowed in the network name"))
            return true
        end

        if fields.name:len() > superquest.config.max_flag_name_length then
            core.chat_send_player(player_name, S("Max flag name length (@1 symbols) exceeded", superquest.config.max_flag_name_length))
            return true
        end

        local new_owner = owner
        if superquest.privileges.can_edit_all(player_name) then
            if not fields.owner then
                return true
            end

            new_owner = fields.owner:gsub('^%s*(.-)%s*$', '%1')

            if new_owner:len() > superquest.config.max_owner_name_length then
                core.chat_send_player(player_name, S("Max owner name length (@1 symbols) exceeded", superquest.config.max_owner_name_length))
                return true
            end
            if new_owner:find("#") then
                core.chat_send_player(player_name, S("Using the '#' symbol is not allowed in the owner name"))
                return true
            end
        end

        local network_obj = superquest.Network(new_owner, new_network)

        local flags_before = string.split(fields.flags_before, "\n")
        if #flags_before > superquest.config.max_flags_before_count then
            core.chat_send_player(player_name, S("Max count of preceding flags (@1) exceeded",
                    superquest.config.max_flags_before_count))
        end
        for _, v in ipairs(flags_before) do
            if v:len() > superquest.config.max_flag_name_length then
                core.chat_send_player(player_name, S("Flag name @1 exceeds the max length", v))
                return true
            end
            if network_obj and network_obj:is_name_or_pos_available(nil, v) then
                core.chat_send_player(player_name, S("Warning: the network @1 doesn't have a flag with the name @2", new_network, v))
            end
        end
        if #flags_before >= 1 then
            meta:set_string("flags_before", core.serialize(flags_before))
        else
            meta:set_string("flags_before", "")
        end

        local old_flag_name = meta:get_string("name")
        local new_flag_name = fields.name
        if old_flag_name == "" then
            old_flag_name = nil
        end
        if new_flag_name == "" then
            new_flag_name = nil
        end

        if new_flag_name and superquest.storage.key_to_coords(new_flag_name) then
            core.chat_send_player(player_name, S("The flag's name shouldn't be in the same format as coordinates"))
            return true
        end

        if network_obj and (old_network ~= new_network or owner ~= new_owner or old_flag_name ~= new_flag_name) then
            if not network_obj:is_name_or_pos_available(pos, new_flag_name) then
                core.chat_send_player(player_name, S("The specified name already belongs to another flag"))
                return true
            end
        end

        if old_network ~= new_network or owner ~= new_owner then
            local old_network_obj = superquest.Network(owner, old_network)
            if old_network_obj then
                old_network_obj:remove_flag(pos, old_flag_name)

                remove_color_mark(pos)
            end

            meta:set_string("owner", new_owner)
            meta:set_string("network", new_network)

            if network_obj then
                network_obj:add_flag(pos, new_flag_name)

                local color = network_obj:get_color()
                if color then
                    local flag_node = core.get_node(pos)
                    add_color_mark(pos, flag_node.param2, color)
                end
            end
        else
            if network_obj and old_flag_name ~= new_flag_name then
                network_obj:set_name_for_flag(pos, old_flag_name, new_flag_name)
            end
        end
        meta:set_string("name", new_flag_name or "")

        if fields.digiline_channel then
            local max_channel_len = superquest.digilines.max_channel_length

            if fields.digiline_channel:len() > max_channel_len then
                core.chat_send_player(player_name, S("Max Digiline channel name length (@1 symbols) exceeded", max_channel_len))
                return true
            end
            meta:set_string("digiline_channel", fields.digiline_channel)
        end

        if superquest.config.flags_autotrigger then
            if not fields.from_coords or not fields.to_coords then
                return true
            end

            local from_coords = fields.from_coords:split(",")
            local to_coords = fields.to_coords:split(",")

            if #from_coords ~= 3 or #to_coords ~= 3 then
                core.chat_send_player(player_name, S("Incorrect format of distance fields"))
                return true
            end

            local from_x = tonumber(from_coords[1])
            local from_y = tonumber(from_coords[2])
            local from_z = tonumber(from_coords[3])
            local to_x = tonumber(to_coords[1])
            local to_y = tonumber(to_coords[2])
            local to_z = tonumber(to_coords[3])

            if not from_x or from_x < 0 or from_x > superquest.config.flags_autotrigger_max_distance or
               not from_y or from_y < 0 or from_y > superquest.config.flags_autotrigger_max_distance or
               not from_z or from_z < 0 or from_z > superquest.config.flags_autotrigger_max_distance or
               not to_x or to_x < 0 or to_x > superquest.config.flags_autotrigger_max_distance or
               not to_y or to_y < 0 or to_y > superquest.config.flags_autotrigger_max_distance or
               not to_z or to_z < 0 or to_z > superquest.config.flags_autotrigger_max_distance then

                core.chat_send_player(player_name, S("Incorrect distance values. Distance values should be integers between @1 and @2",
                    0, superquest.config.flags_autotrigger_max_distance))
                return true
            end

            meta:set_int("scan_from_x", from_x)
            meta:set_int("scan_from_y", from_y)
            meta:set_int("scan_from_z", from_z)
            meta:set_int("scan_to_x", to_x)
            meta:set_int("scan_to_y", to_y)
            meta:set_int("scan_to_z", to_z)
        end

        if superquest.config.teleportation then
            if not fields.teleport_to then
                return true
            end

            if fields.teleport_to ~= "" then
                local teleport_to = fields.teleport_to:split(",")
                if #teleport_to ~= 3 then
                    core.chat_send_player(player_name, S("Incorrect format of distance fields"))
                    return true
                end

                local tp_to_x = tonumber(teleport_to[1])
                local tp_to_y = tonumber(teleport_to[2])
                local tp_to_z = tonumber(teleport_to[3])

                local max_dist = superquest.config.flags_teleport_max_distance

                if not tp_to_x or tp_to_x < -max_dist or tp_to_x > max_dist or
                   not tp_to_y or tp_to_y < -max_dist or tp_to_y > max_dist or
                   not tp_to_z or tp_to_z < -max_dist or tp_to_z > max_dist then
                    core.chat_send_player(player_name, S("Incorrect distance values. Distance values should be integers between @1 and @2", -max_dist, max_dist))
                    return true
                end

                meta:set_string("teleport_to", tostring(tp_to_x)..","..tostring(tp_to_y)..","..tostring(tp_to_z))
            else
                meta:set_string("teleport_to", "")
            end
        end
    elseif fields.digiline_guide then
        superquest.digilines.open_guide_formspec(player_name)
    end

    return true
end)

local flag_nodes = { "superquest:flag" }
if digilines then
    table.insert(flag_nodes, "superquest:flag_digiline_requiring")
end

if superquest.config.flags_autotrigger then
    core.register_abm({
        nodenames = flag_nodes,
        interval = superquest.config.flags_autotrigger_interval,
        chance = 1,
        action = function(pos, node)
            local meta = core.get_meta(pos)
            local from_x_dev = meta:get_int("scan_from_x")
            local from_y_dev = meta:get_int("scan_from_y")
            local from_z_dev = meta:get_int("scan_from_z")
            local to_x_dev = meta:get_int("scan_to_x")
            local to_y_dev = meta:get_int("scan_to_y")
            local to_z_dev = meta:get_int("scan_to_z")

            local from_v = vector.subtract(pos, vector.new(from_x_dev, from_y_dev, from_z_dev))
            local to_v = vector.add(pos, vector.new(to_x_dev, to_y_dev, to_z_dev))

            local objs = core.get_objects_in_area(from_v, to_v)

            for _, obj in pairs(objs) do
                if obj:is_player() then
                    try_to_count_flag(pos, obj, false, node.name == "superquest:flag_digiline_requiring")
                end
            end
        end
    })
end

core.register_lbm({
    name = "superquest:set_color_flag",
    nodenames = flag_nodes,
    run_at_every_load = true,

    action = function(pos, node)
        local meta = core.get_meta(pos)
        local owner = meta:get_string("owner")
        local network = meta:get_string("network")

        local network_obj = superquest.Network(owner, network)
        if not network_obj then
            return
        end

        local color = network_obj:get_color()
        if color then
            add_color_mark(pos, node.param2, color)
        end
    end
})

superquest.flag = {}

superquest.flag.is_flag = function(node_name)
    return node_name == "superquest:flag" or node_name == "superquest:flag_digiline_requiring"
end

superquest.flag.add_color_mark = add_color_mark
superquest.flag.remove_color_mark = remove_color_mark

if superquest.config.teleportation then
    superquest.flag.get_tp_destination = function(owner, network_name, pos)
        local node = superquest.utils.get_node(pos)
        if not superquest.flag.is_flag(node.name) then
            return nil
        end

        local meta = core.get_meta(pos)
        if meta:get_string("owner") ~= owner or meta:get_string("network") ~= network_name then
            return nil
        end

        local teleport_to = meta:get_string("teleport_to")
        if teleport_to == "" then
            return nil
        end

        local teleport_to = teleport_to:split(",")

        return pos + vector.new(teleport_to[1], teleport_to[2], teleport_to[3])
    end

    core.register_on_dieplayer(function(player)
        local player_name = player:get_player_name()
        local allowed = superquest.teleportation.allow_tp_for_flag(player_name)
        if allowed then
            core.chat_send_player(player_name, S("You've just died!").." "..S("You can type '/superquest teleport' to teleport to the last triggered flag."))
        end
    end)
end

if digilines then
    superquest.flag.remove_allowed_players = allowed_players_remove_for_all
end
